var isArray = Array.isArray;

var EventEmitter = function EventEmitter() {
    "use strict";
    this._filters = {};
};

exports.EventEmitter = EventEmitter;
var defaultMaxListeners = 10;

EventEmitter.prototype.setMaxListeners = function(nb) {
    "use strict";
    if (!this._events) {
        this._events = {};
    }
    this._maxListeners = nb;
};

EventEmitter.prototype.emit = function emit() {
    "use strict";
    var type = arguments[0];
    if (type === 'error') {
        if (!this._events || !this._events.error ||
        isArray(this._events.error) && !this._events.error.length) {
            if (arguments[1] instanceof Error) {
                throw arguments[1]; 
            } else {
                throw new CasperError("Uncaught, unspecified 'error' event.");
            }
        }
    }

    if (!this._events) {
        return false;
    }
    var handler = this._events[type];
    if (!handler) {
        return false;
    }

    if (typeof handler === 'function') {
        try {
            switch (arguments.length) {
            case 1:
                handler.call(this);
                break;
            case 2:
                handler.call(this, arguments[1]);
                break;
            case 3:
                handler.call(this, arguments[1], arguments[2]);
                break;
            default:
                var l = arguments.length;
                var args = new Array(l - 1);
                for (var i = 1; i < l; i++) {
                    args[i - 1] = arguments[i];
                }
                handler.apply(this, args);
            }
        } catch (err) {
            this.emit('event.error', err);
        }
        return true;

    } else if (isArray(handler)) {
        var l = arguments.length;
        var args = new Array(l - 1);
        for (var i = 1; i < l; i++) {
            args[i - 1] = arguments[i];
        }

        var listeners = handler.slice();
        for (var i = 0, l = listeners.length; i < l; i++) {
            listeners[i].apply(this, args);
        }
        return true;

    } else {
        return false;
    }
};
EventEmitter.prototype.addListener = function addListener(type, listener) {
    "use strict";
    if (typeof listener !== 'function') {
        throw new CasperError('addListener only takes instances of Function');
    }

    if (!this._events) {
        this._events = {};
    }
    this.emit('newListener', type, listener);

    if (!this._events[type]) {
        this._events[type] = listener;
    } else if (isArray(this._events[type])) {
        this._events[type][ type === 'fail' ? 'unshift' : 'push'](listener);
        if (!this._events[type].warned) {
            var m;
            if (typeof this._maxListeners !== 'undefined') {
                m = this._maxListeners;
            } else {
                m = defaultMaxListeners;
            }

            if (m && m > 0 && this._events[type].length > m) {
                this._events[type].warned = true;
                console.error('(node) warning: possible EventEmitter memory ' +
                            'leak detected. %d listeners added. ' +
                            'Use emitter.setMaxListeners() to increase limit.',
                            this._events[type].length);
                console.trace();
            }
        }
    } else {
        this._events[type] = type === 'fail' ? [listener, this._events[type]] : [this._events[type], listener];
    }
    return this;
};
EventEmitter.prototype.prependListener = function prependListener(type, listener) {
    "use strict";
    if (typeof listener !== 'function') {
        throw new CasperError('addListener only takes instances of Function');
    }

    if (!this._events) {
        this._events = {};
    }
    this.emit('newListener', type, listener);

    if (!this._events[type]) {
        this._events[type] = listener;
    } else if (isArray(this._events[type])) {
        this._events[type].unshift(listener);
        if (!this._events[type].warned) {
            var m;
            if (typeof this._maxListeners !== 'undefined') {
                m = this._maxListeners;
            } else {
                m = defaultMaxListeners;
            }

            if (m && m > 0 && this._events[type].length > m) {
                this._events[type].warned = true;
                console.error('(node) warning: possible EventEmitter memory ' +
                            'leak detected. %d listeners added. ' +
                            'Use emitter.setMaxListeners() to increase limit.',
                            this._events[type].length);
                console.trace();
            }
        }
    } else {
        this._events[type] = [listener, this._events[type]];
    }
    return this;
};
EventEmitter.prototype.on = EventEmitter.prototype.addListener;

EventEmitter.prototype.once = function once(type, listener) {
    "use strict";
    if (typeof listener !== 'function') {
        throw new CasperError('.once only takes instances of Function');
    }

    var self = this;
    var g = function g() {
        self.removeListener(type, g);
        listener.apply(this, arguments);
    };

    g.listener = listener;
    self.on(type, g);

    return this;
};
EventEmitter.prototype.removeListener = function removeListener(type, listener) {
    "use strict";
    if (typeof listener !== 'function') {
        throw new CasperError('removeListener only takes instances of Function');
    }

    if (!this._events || !this._events[type]) {
        return this;
    }

    var list = this._events[type];

    if (isArray(list)) {
        var position = -1;
        for (var i = 0, length = list.length; i < length; i++) {
            if (list[i] === listener ||
                list[i].listener && list[i].listener === listener) {
                position = i;
                break;
            }
        }
        if (position < 0) {
            return this;
        }
        list.splice(position, 1);
        if (list.length === 0) {
            delete this._events[type];
        }
    } else if (list === listener ||
             list.listener && list.listener === listener) {
        delete this._events[type];
    }

    return this;
};
EventEmitter.prototype.removeAllListeners = function removeAllListeners(type) {
    "use strict";
    if (arguments.length === 0) {
        this._events = {};
        return this;
    }
    if (type && this._events && this._events[type]) {
        this._events[type] = null;
    }
    return this;
};
EventEmitter.prototype.listeners = function listeners(type) {
    "use strict";
    if (!this._events) {
        this._events = {};
    }
    if (!this._events[type]) {
        this._events[type] = [];
    }
    if (!isArray(this._events[type])) {
        this._events[type] = [this._events[type]];
    }
    return this._events[type];
};
EventEmitter.prototype.filter = function filter() {
    "use strict";
    var type = arguments[0], res;
    if (!this._filters) {
        this._filters = {};
    }
    var _filter = this._filters[type];
    if (typeof _filter === 'function') {
        res = _filter.apply(this, Array.prototype.slice.call(arguments, 1));
    }
    this.emit(Array.prototype.slice.call(arguments, 0).concat([res]));
    return res;
};

EventEmitter.prototype.removeAllFilters = function removeAllFilters(type) {
    "use strict";
    if (arguments.length === 0) {
        this._filters = {};
        return this;
    }
    if (type && this._filters && this._filters[type]) {
        this._filters[type] = null;
    }
    return this;
};
EventEmitter.prototype.setFilter = function setFilter(type, filterFn) {
    "use strict";
    if (!this._filters) {
        this._filters = {};
    }
    if (typeof filterFn !== 'function') {
        throw new CasperError('setFilter only takes instances of Function');
    }
    if (!this._filters[type]) {
        this._filters[type] = filterFn;
        return true;
    }
    return false;
};
